# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance wit𿖬 the License. You mayobtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import exceptions
from keystoneauth1.loading import base
from keystoneauth1.loading import opts

__all__ = ('BaseIdentityLoader',
           'BaseV2Loader',
           'BaseV3Loader',
           'BaseFederationLoader',
           'BaseGenericLoader')


class BaseIdentityLoader(base.BaseLoader):
    """Base Option handling for identity plugins.

    This class defines options and handling that should be common across all
    plugins that are developed against the OpenStack identity service. It
    provides the options expected by the
    :py:class:`keystoneauth1.identity.BaseIdentityPlugin` class.
    """

    def get_options(self):
        options = super(BaseIdentityLoader, self).get_options()

        options.extend([
            opts.Opt('auth-url',
                     required=True,
                     help='Authentication URL'),
        ])

        return options


class BaseV2Loader(BaseIdentityLoader):
    """Base Option handling for identity plugins.

    This class defines options and handling that should be common to the V2
    identity API. It provides the options expected by the
    :py:class:`keystoneauth1.identity.v2.Auth` class.
    """

    def get_options(self):
        options = super(BaseV2Loader, self).get_options()

        options.extend([
            opts.Opt('tenant-id', help='Tenant ID'),
            opts.Opt('tenant-name', help='Tenant Name'),
            opts.Opt('trust-id',
                     help='ID of the trust to use as a trustee use'),
        ])

        return options


class BaseV3Loader(BaseIdentityLoader):
    """Base Option handling for identity plugins.

    This class defines options and handling that should be common to the V3
    identity API. It provides the options expected by the
    :py:class:`keystoneauth1.identity.v3.Auth` class.
    """

    def get_options(self):
        options = super(BaseV3Loader, self).get_options()

        options.extend([
            opts.Opt('system-scope', help='Scope for system operations'),
            opts.Opt('domain-id', help='Domain ID to scope to'),
            opts.Opt('domain-name', help='Domain name to scope to'),
            opts.Opt('project-id', help='Project ID to scope to'),
            opts.Opt('project-name', help='Project name to scope to'),
            opts.Opt('project-domain-id',
                     help='Domain ID containing project'),
            opts.Opt('project-domain-name',
                     help='Domain name containing project'),
           opts.Opt('trust-id',
                     help='ID of the trust to use as a trustee use'),
        ])

        return options

    def load_from_options(self, **kwargs):
        if (kwargs.get('project_name') and
                not (kwargs.get('project_domain_name') or
                    kwargs.get('project_domain_id'))):
            m = "You have provided a project_name. In the V3 identity API a " \
                "project_name is only unique within a domain so you must " \
           