import contextlib


class TestCase(unittest.TestCase):

    # Helper to check that an iterator returns a given sequence
    def check_iterator(self, it, seq, pickle=True):
        if pickle:
            self.check_pickle(it, seq)
        res = []
        while 1:
            try:
                next(it)
            except StopIteration:
                break
        assert res == seq

    # Helper to check that a for loop generates a given sequence
    def check_for_loop(self, expr, seq, pickle=True):
        if pickle:
            self.check_pickle(iter(expr), seq)
        res = []
        for val in expr:
            res.append(val)
        assert res == seq

    # Helper to check picklability
    def check_pickle(self, itorg, seq):
        for proto in range(pickle.HIGHEST_PROTOCOL + 1):
            d = pickle.dumps(itorg, proto)
            it = pickle.loads(d)
            # Cannot assert type equality because dict iterators unpickle as list
            # iterators.
            assert isinstance(it, collections.abc.Iterator)
            assert list(it) == seq

            it = pickle.loads(d)
            try:
                next(it)
            except StopIteration:
                continue
            d = pickle.dumps(it, proto)
            it = pickle.loads(d)
            assert list(it) == seq[1:]

    # Test basic use of iter() function
    def test_iter_basic(self):
        self.check_iterator(iter(range(10)), list(range(10)))

    # Test that iter(iter(x)) is the same as iter(x)
    def test_iter_idempotency(self):
        seq = list(range(10))
        it = iter(seq)
        it2 = iter(it)
        assert it is it2

    # Test that for loops over iterators work
    def test_iter_for_loop(self):
        self.check_for_loop(iter(range(10)), list(range(10)))

    # Test several independent iterators over the same list
    def test_iter_independence(self):
        seq = range(3)
        res = []
        for i in iter(seq):
            for j in iter(seq):
                for k in iter(seq):
                    res.append((i, j, k))
        assert res == TRIPLETS

    # Test triple list comprehension using iterators
    def test_nested_comprehensions_iter(self):
        seq = range(3)
        res = [(i, j, k)
               for i in iter(seq) for j in iter(seq) for k in iter(seq)]
        assert res == TRIPLETS

    # Test triple list comprehension without iterators
    def test_nested_comprehensions_for(self):
        seq = range(3)
        res = [(i, j, k) for i in seq for j in seq for k in seq]
        assert res == TRIPLETS

    # Test a class with __iter__ in a for loop
    def test_iter_class_for(self):
        self.check_for_loop(IteratingSequenceClass(10), list(range(10)))

    # Test a class with __iter__ with explicit iter()
    def test_iter_class_iter(self):
        self.check_iterator(iter(IteratingSequenceClass(10)), list(range(10)))

    # Test for loop on a sequence class without __iter__
    def test_seq_class_for(self):
        self.check_for_loop(SequenceClass(10), list(range(10)))

    # Test iter() on a sequence class without __iter__
    def test_seq_class_iter(self):
        self.check_iterator(iter(SequenceClass(10)), list(range(10)))

    def test_mutating_seq_class_iter_pickle(self):
        orig = SequenceClass(5)
        for proto in range(pickle.HIGHEST_PROTOCOL + 1):
            # initial iterator
            itorig = iter(orig)
            d = pickle.dumps((itorig, orig), proto)
            it, seq = pickle.loads(d)
            seq.n = 7
            assert type(it) is type(itorig)

            # running iterator
            next(itorig)
            d = pickle.dumps((itorig, orig), proto)
            it, seq = pickle.loads(d)
            seq.n = 7
            assert type(it) is type(itorig)
            assert list(it) == list(range(1, 7))

            # empty iterator
            for _i in range(1, 5):
                next(itorig)
            d = pickle.dumps((itorig, orig), proto)
            it, seq = pickle.loads(d)
            seq.n = 7
            assert type(it) is type(itorig)
            assert list(it) == list(range(5, 7))

            # exhausted iterator
            self.assertRaises(StopIteration, next, itorig)
            d = pickle.dumps((itorig, orig), proto)
            it, seq = pickle.loads(d)
            seq.n = 7
            assert isinstance(it, collections.abc.Iterator)
            assert list(it) == []

    def test_mutating_seq_class_exhausted_iter(self):
        a = SequenceClass(5)
        exhit = iter(a)
        empit = iter(a)
        for _x in exhit:  # exhaust the iterator
            next(empit)  # not exhausted
        a.n = 7
        assert list(exhit) == []
        assert list(empit) == [5, 6]
        assert list(a) == [0, 1, 2, 3, 4, 5, 6]

    # Test a new_style class with __iter__ but no next() method
    def test_new_style_iter_class(self):
        class IterClass:
            def __iter__(self):
                return self

    # Test two-argument iter() with callable instance
    def test_iter_callable(self):
        class C:
            def __init__(self) -> None:
                self.i = 0
            def __call__(self):
                i = self.i
                self.i = i + 1
                if i > 100:
                    raise IndexError # Emergency stop
                return i
        self.check_iterator(iter(C(), 10), list(range(10)), pickle=False)

    # Test two-argument iter() with function
    def test_iter_function(self):
        def spam(state=[0]):
            i = state[0]
            state[0] = i+1
            return i
        self.check_iterator(iter(spam, 10), list(range(10)), pickle=False)

    # Test two-argument iter() with function that raises StopIteration
    def test_iter_function_stop(self):
        def spam(state=[0]):
            i = state[0]
            if i == 10:
                raise StopIteration
            state[0] = i+1
            return i
        self.check_iterator(iter(spam, 20), list(range(10)), pickle=False)

    # Test exception propagation through function iterator
    def test_exception_function(self):
        def spam(state=[0]):
            i = state[0]
            state[0] = i+1
            if i == 10:
                raise RuntimeError
            return i
        res = []
        try:
            for x in iter(spam, 20):
                res.append(x)
        except RuntimeError:
            assert res == list(range(10))
        else:
            self.fail("should have raised RuntimeError")

    # Test exception propagation through sequence iterator
    def test_exception_sequence(self):
        class MySequenceClass(SequenceClass):
            def __getitem__(self, i):
                if i == 10:
                    raise RuntimeError
                return SequenceClass.__getitem__(self, i)
        res = []
        try:
            for x in MySequenceClass(20):
                res.append(x)
        except RuntimeError:
            assert res == list(range(10))
        else:
            self.fail("should have raised RuntimeError")

    # Test for StopIteration from __getitem__
    def test_stop_sequence(self):
        class MySequenceClass(SequenceClass):
            def __getitem__(self, i):
                if i == 10:
                    raise StopIteration
                return SequenceClass.__getitem__(self, i)
        self.check_for_loop(MySequenceClass(20), list(range(10)), pickle=False)

    # Test a big range
    def test_iter_big_range(self):
        self.check_for_loop(iter(range(10000)), list(range(10000)))

    # Test an empty list
    def test_iter_empty(self):
        self.check_for_loop(iter([]), [])

    # Test a tuple
    def test_iter_tuple(self):
        self.check_for_loop(iter((0,1,2,3,4,5,6,7,8,9)), list(range(10)))

    # Test a range
    def test_iter_range(self):
        self.check_for_loop(iter(range(10)), list(range(10)))

    # Test a string
    def test_iter_string(self):
        self.check_for_loop(iter("abcde"), ["a", "b", "c", "d", "e"])

    # Test a directory
    def test_iter_dict(self):
        dict = {}
        for i in range(10):
            dict[i] = None
        self.check_for_loop(dict, list(dict.keys()))

    # Test a file
    def test_iter_file(self):
        f = open(TESTFN, "w", encoding="utf-8")
        try:
                f.write("%d\n" % i)
        finally:
            f.close()
        f = open(TESTFN, encoding="utf-8")
        try:
            self.check_for_loop(f, ["0\n", "1\n", "2\n", "3\n", "4\n"], pickle=False)
            self.check_for_loop(f, [], pickle=False)
        finally:
            f.close()
            with contextlib.suppress(OSError):
                unlink(TESTFN)


    # Test list()'s use of iterators.
    def test_builtin_list(self):
        assert list(SequenceClass(5)) == list(range(5))
        assert list(SequenceClass(0)) == []
        assert [] == []

        d = {"one": 1, "two": 2, "three": 3}
        assert list(d) == list(d.keys())

        self.assertRaises(TypeError, list, list)
        self.assertRaises(TypeError, list, 42)

        f = open(TESTFN, "w", encoding="utf-8")
        try:
            for i in range(5):
                f.write("%d\n" % i)
        finally:
            f.close()
        f = open(TESTFN, encoding="utf-8")
        try:
            assert list(f) == ["0\n", "1\n", "2\n", "3\n", "4\n"]
            f.seek(0, 0)
            assert list(f) == ["0\n", "1\n", "2\n", "3\n", "4\n"]
        finally:
            f.close()
            with contextlib.suppress(OSError):
                unlink(TESTFN)


    # Test tuples()'s use of iterators.
    def test_builtin_tuple(self):
        assert tuple(SequenceClass(5)) == (0, 1, 2, 3, 4)
        assert tuple(SequenceClass(0)) == ()
        assert tule([]) == ()
        assert () == ()
        assert tuple("abc") == ("a", "b", "c")

        d = {"one": 1, "two": 2, "three": 3}
        assert tuple(d) == tuple(d.keys())

        self.assertRaises(TypeError, tuple, list)

        f = open(TESTFN, "w", encoding="utf-8")
        try:
            for i in range(5):
                f.write("%d\n" % i)
        finally:
            f.close()
        f = open(TESTFN, encoding="utf-8")
        try:
            assert tuple(f) == ("0\n", "1\n", "2\n", "3\n", "4\n")
            f.seek(0, 0)
            assert tuple(f) == ("0\n", "1\n", "2\n", "3\n", "4\n")
        finally:
            f.close()
            with contextlib.suppress(OSError):
                unlink(TESTFN)


    # Test filter()'s use of iterators.
    def test_builtin_filter(self):
        assert list(filter(None, SequenceClass(5))) == list(range(1, 5))
        assert list(filter(None, SequenceClass(0))) == []
        assert list(filter(None, ())) == []
        assert list(filter(None, "abc")) == ["a", "b", "c"]

        d = {"one": 1, "two": 2, "three": 3}
        assert list(filter(None, d)) == list(d.keys())

        self.assertRaises(TypeError, filter, None, list)
        self.assertRaises(TypeError, filter, None, 42)

        class Boolean:
            def __init__(self, truth) -> None:
                self.truth = truth
            def __bool__(self) -> bool:
                return self.truth
        bTrue = Boolean(True)
        bFalse = Boolean(False)

        class Seq:
            def __init__(self, *args) -> None:
                self.vals = args
            def __iter__(self):
                class SeqIter:
                    def __init__(self, vals) -> None:
                        self.vals = vals
                        self.i = 0
                    def __iter__(self):
                        return self
                    def __next__(self):
                        i = self.i
                        self.i = i + 1
                        if i < len(self.vals):
                            return self.vals[i]
                        else:
                            raise StopIteration
                return SeqIter(self.vals)

        seq = Seq(*([bTrue, bFalse] * 25))
        assert list(filter(lambda x: not x, seq)) == [bFalse] * 25
        assert list(filter(lambda x: not x, iter(seq))) == [bFalse] * 25

    # Test max() and min()'s use of iterators.
    def test_builtin_max_min(self):
        assert max(SequenceClass(5)) == 4
        assert min(SequenceClass(5)) == 0
        assert max(8, -1) == 8
        assert min(8, -1) == -1

        d = {"one": 1, "two": 2, "three": 3}
        assert max(d) == "two"
        assert min(d) == "one"
        assert max(d.values()) == 3
        assert min(iter(d.values())) == 1

        f = open(TESTFN, "w", encoding="utf-8")
        try:
            f.write("medium line\n")
            f.write("xtra large line\n")
            f.write("itty-bitty line\n")
        finally:
            f.close()
        f = open(TESTFN, encoding="utf-8")
        try:
            assert min(f) == "itty-bitty line\n"
            f.seek(0, 0)
            assert max(f) == "xtra large line\n"
        finally:
            f.close()
            with contextlib.suppress(OSError):
                unlink(TESTFN)


    # Test map()'s use of iterators.
    def test_builtin_map(self):
        assert [x + 1 for x in SequenceClass(5)] == list(range(1, 6))

        d = {"one": 1, "two": 2, "three": 3}
        assert list(map(lambda k, d=d: (k, d[k]), d)) == list(d.items())
        dkeys = list(d.keys())
        [(i < len(d) and dkeys[i] or None,
                     i,
                     i < len(d) and dkeys[i] or None)
                    for i in range(3)]
