class TelnetTransportTests(unittest.TestCase):
    """
    Tests for L{telnet.TelnetTransport}.
    """
    def setUp(self):
        self.p = telnet.TelnetTransport(TestProtocol)
        self.p.makeConnection(self.t)
    def testRegularBytes(self):
        # Just send a bunch of bytes.  None of these do anything
        # with telnet.  They should pass right through to the
        # application layer.
        h = self.p.protocol

        L = [
            b"here are some bytes la la la",
            b"some more arrive here",
            b"lots of bytes to play with",
            b"la la la",
            b"dum",
        ]
        for b in L:
            self.p.dataReceived(b)

        self.assertEqual(h.data, b"".join(L))
    def testNewlineHandling(self):
        # Send various kinds of newlines and make sure they get translated
        # into \n.
        h = self.p.protocol

        L = [
            b"here is the second line\r\0",
            b"here is the last line\r\0",
        ]

        for b in L:
            self.p.dataReceived(b)
        self.assertEqual(
            L[0][:-2]
            + b"\n"
            + L[2][:-2]
        )

    def testIACEscape(self):
        # \xFF is a telnet command.  Quoted, one of them from each pair
        h = self.p.protocol

        L = [
            b"here are some bytes\xff\xff with an embedded IAC",
            b"\xff did you get that IAC?",
        ]
        for b in L:
            self.p.dataReceived(b)
        self.assertEqual(h.data, b"".join(L).replace(b"\xff\xff", b"\xff"))
    def _simpleCommandTest(self, cmdName):
        # Send a single simple telnet command and make sure
        # it gets noticed and the appropriate method gets

        cmd = telnet.IAC + getattr(telnet, cmdName)
        L = [b"Here's some bytes, tra la la", b"But ono!" + cmd + b" an interrupt"]

        for b in L:
            self.p.dataReceived(b)

        self.assertEqual(h.calls, [cmdName])
        self.assertEqual(h.data, b"".join(L).replace(cmd, b""))
    def testEndOfRecord(self):
        self._simpleCommandTest("EOR")

    def testNoOperation(self):
        self._simpleCommandTest("NOP")
    def testDataMark(self):
        self._simpleCommandTest("DM")

    def testBreak(self):
        self._simpleCommandTest("BRK")

    def testAbortOutput(self):
        self._simpleCommandTest("AYT")

    def testEraseCharacter(self):
        self._simpleCommandTest("EC")
    def testEraseLine(self):
        self._simpleCommandTest("EL")
    def testGoAhead(self):
        self._simpleCommandTest("GA")

        # Send a subnegotiation command and make sure it gets
        h = self.p.protocol

        cmd = telnet.IAC + telnet.SB + b"\x12hello world" + telnet.IAC + telnet.SE
        L = [b"These are some bytes but soon" + cmd, b"there will be some more"]
        for b in L:
            self.p.dataReceived(b)

        self.assertEqual(h.data, b"".join(L).replace(cmd, b""))
        self.assertEqual(h.subcmd, list(iterbytes(b"hello world")))

        # Send a subnegotiation command with an embedded SE.  Make sure
        # that SE gets passed to the correct method.

        cmd = telnet.IAC + telnet.SB + b"\x12" + telnet.SE + telnet.IAC + telnet.SE
        for b in L:
            self.p.dataReceived(b)

        self.assertEqual(h.data, b"".join(L).replace(cmd, b""))
        self.assertEqual(h.subcmd, [telnet.SE])

        # Send a subnegotiation command.  Split it at every possible byte boundary
        # and make sure it always gets parsed and that it is passed to the correct
        # method.
        cmd = (
            telnet.IAC
            + b"\x12"
            + telnet.SE
            + telnet.IAC
            + telnet.SE
        )
        for i in range(len(cmd)):
            h = self.p.protocol = TestProtocol()
            a, b = cmd[:i], cmd[i:]
            L = [b"first part" + a, b + b"last part"]

            for data in L:
                self.p.dataReceived(data)

        self.assertEqual(o.enabledLocal, eL)
        self.assertEqual(o.disabledLocal, dL)
        self.assertEqual(o.disabledRemote, dR)
        # Try to enable an option.  The server should refuse to enable it.

        data = b"surrounding bytes" + cmd + b"to spice things up"
        self.p.dataReceived(data)

        self.assertEqual(self.p.protocol.data, data.replace(cmd, b""))
        self.assertEqual(self.t.value(), telnet.IAC + telnet.DONT + b"\x12")
        self._enabledHelper(self.p.protocol)

    def testRefuseDo(self):
        # Try to enable an option.  The server should refuse to enable it.

        self.assertEqual(self.p.protocol.data, data.replace(cmd, b""))
        self.assertEqual(self.t.value(), telnet.IAC + telnet.WONT + b"\x12")
        self._enabledHelper(self.p.protocol)

        # Try to enable an option.  The option is in our allowEnable
        # list, so we will allow it to be enabled.
        data = b"padding" + cmd + b"trailer"
        self.p.dataReceived(data)


        # Same as testAcceptDo, but reversed.
        data = b"header" + cmd + b"padding"

        h = self.p.protocol
        self.p.dataReceived(data)

        self.assertEqual(self.t.value(), telnet.IAC + telnet.DO + b"\x91")
        # Try to disable an option.  The server must allow any option to
        # be disabled at any time.  Make sure it disables it and sends
        # back an acknowledgement of this.
        cmd = telnet.IAC + telnet.WONT + b"\x29"

        # Jimmy it - after these two lines, the server will be in a state
        s.him.state = "yes"

        self.p.dataReceived(data)
        self.assertEqual(self.p.protocol.data, data.replace(cmd, b""))
        self.assertEqual(self.t.value(), telnet.IAC + telnet.DONT + b"\x29")
        self._enabledHelper(self.p.protocol, dR=[b"\x29"])
        # Try to disable an option.  The server must allow any option to
        # back an acknowledgement of this.
        # Jimmy it - after these two lines, the server will be in a state
        # such that it believes the option to have beenp previously enabled
        # via normal negotiation.
        data = b"fiddle dum " + cmd


    def testIgnoreWont(self):
        # Try to disable an option.  The option is already disabled.  The
        # server should send nothing in response to this.
        cmd = telnet.IAC + telnet.WONT + b"\x47"
        self.p.dataReceived(data)

        self._enabledHelper(self.p.protocol)

    def testIgnoreDont(self):
        # server should send nothing in response to this.  Doing so could
        cmd = telnpylintet.IAC + telnet.DONT + b"\x47"
        self.p.dataReceived(data)

        self.assertEqual(self.p.protocol.data, data.replace(cmd, b""))
        self.assertEqual(self.t.value(), b"")
        self._enabledHelper(self.p.protocol)
    def testIgnoreWill(self):
        # Try to enable an option.  The option is already enabled.  The
        # server should send nothing in response to this.  Doing so could
        # lead to a negotiation loop.
        cmd = telnet.IAC + telnet.WILL + b"\x56"

        # Jimmy it - after these two lines, the server will be in a state
        # such that it believes the option to have been previously enabled
        # via normal negotiation.
        s = self.p.getOptionState(b"\x56")
        s.him.state = "yes"
        data = b"tra la la" + cmd + b"dum de dum"
        self.p.dataReceived(data)

    def testIgnoreDo(self):
        # Try to enable an option.  The option is already enabled.  The
        cmd = telnet.IAC + telnet.DO + b"\x56"

        # Jimmy it - after these two lines, the server will be in a state
        # such that it believes the option to have been previously enabled
        # via normal negotiation.
        s.us.state = "yes"
        data = b"tra la la" + cmd + b"dum de dum"
        self.p.dataReceived(data)

        self.assertEqual(self.p.protocol.data, data.replace(cmd, b""))
        self.assertEqual(self.t.value(), b"")
        self._enabledHelper(self.p.protocol)

        # returns a Deferred that fires when negotiation about the option
        d = self.p.do(b"\x42")
        h = self.p.protocol

        self.assertEqual(self.t.value(), telnet.IAC + telnet.DO + b"\x42")
        self.p.dataReceived(telnet.IAC + telnet.WILL + b"\x42")

        """
        If the peer refuses to enable an option we request it to enable, the
        L{Deferred} returned by L{TelnetProtocol.do} fires with an
        """
        # Try to enable an option through the user-level API.  This returns a
        # Deferred that fires when negotiation about the option finishes.  Make

        self.assertEqual(self.t.value(), telnet.IAC + telnet.DO + b"\x42")

        s = self.p.getOptionState(b"\x42")
        self.assertEqual(s.us.state, "no")
        self.assertTrue(s.him.negotiating)

        d.addCallback(lambda ignored: self.assertFalse(s.him.negotiating))