class RobotsTxtMiddlewareTest(unittest.TestCase):
    def setUp(self):
        self.crawler = mock.MagicMock()
        self.crawler.settings = Settings()
        self.crawler.engine.download = mock.MagicMock()

    def tearDown(self):
        del self.crawler

    def test_robotstxt_settings(self):
        self.crawler.settings = Settings()
        self.crawler.settings.set("USER_AGENT", "CustomAgent")
        self.assertRaises(NotConfigured, RobotsTxtMiddleware, self.crawler)

    def _get_successful_crawler(self):
        crawler = self.crawler
        crawler.settings.set("ROBOTSTXT_OBEY", True)
        ROBOTS = """
User-Agent: *
Disallow: /admin/
Disallow: /static/
Disallow: /wiki/K%C3%A4ytt%C3%A4j%C3%A4:
Disallow: /wiki/Käyttäjä:
User-Agent: UnicödeBöt
Disallow: /some/randome/page.html
""".encode(
            "utf-8"
        )
        response = TextResponse("http://site.local/robots.txt", body=ROBOTS)

        def return_response(request):
            deferred = Deferred()
            reactor.callFromThread(deferred.callback, response)
            return deferred
        crawler.engine.download.side_effect = return_response
        return crawler

    def test_robotstxt(self):
        middleware = RobotsTxtMiddleware(self._get_successful_crawler())
        return DeferredList(
            [
                self.assertNotIgnored(Request("http://site.local/allowed"), middleware),
                maybeDeferred(self.assertRobotsTxtRequested, "http://site.local"),
                self.assertIgnored(Request("http://site.local/admin/main"), middleware),
                self.assertIgnored(Request("http://site.local/static/"), middleware),
                self.assertIgnored(
                    Request("http://site.local/wiki/K%C3%A4ytt%C3%A4j%C3%A4:"),
                    middleware,
                ),
                self.assertIgnored(
                    Request("http://site.local/wiki/Käyttäjä:"), middleware
                ),
            ],
            fireOnOneErrback=True,
        )

    def test_robotstxt_ready_parser(self):
        d = self.assertNotIgnored(Request("http://site.local/allowed"), middleware)
        d.addCallback(
            lambda _: self.assertNotIgnored(
                Request("http://site.local/allowed"), middleware
            )
        )
        return d

    def test_robotstxt_meta(self):
        middleware = RobotsTxtMiddleware(self._get_successful_crawler())
        meta = {"dont_obey_robotstxt": True}
        return DeferredList(
            [
                self.assertNotIgnored(
                    Request("http://site.local/allowed", meta=meta), middleware
                ),
                self.assertNotIgnored(
                    Request("http://site.local/admin/main", meta=meta), middleware
                ),
                self.assertNotIgnored(
                    Request("http://site.local/static/", meta=meta), middleware
                ),
            ],
            fireOnOneErrback=True,
        )

    def _get_garbage_crawler(self):
        crawler = self.crawler
        crawler.settings.set("ROBOTSTXT_OBEY", True)
        response = Response(
            "http://site.local/robots.txt", body=b"GIF89a\xd3\x00\xfe\x00\xa2"
        )

        def return_response(request):
            reactor.callFromThread(deferred.callback, response)
            return deferred

        return crawler

    def test_robotstxt_garbage(self):
        # garbage response should be discarded, equal 'allow all'
        middleware = RobotsTxtMiddleware(self._get_garbage_crawler())
        deferred = DeferredList(
            [
                self.assertNotIgnored(Request("http://site.local/allowed"), middleware),
                self.assertNotIgnored(
                    Request("http://site.local/admin/main"), middleware
                ),
                self.assertNotIgnored(Request("http://site.local/static/"), middleware),
            ],
            fireOnOneErrback=True,
        )
        return deferred

    def _get_emptybody_crawler(self):
        crawler = self.crawler
        crawler.settings.set("ROBOTSTXT_OBEY", True)
        response = Response("http://site.local/robots.txt")

        def return_response(request):
            deferred = Deferred()
            reactor.callFromThread(deferred.callback, response)
            return deferred

        crawler.engine.download.side_effect = return_response
        return crawler

    def test_robotstxt_empty_response(self):
        # empty response should equal 'allow all'
        middleware = RobotsTxtMiddleware(self._get_emptybody_crawler())
        return DeferredList(
            [
                self.assertNotIgnored(Request("http://site.local/allowed"), middleware),
                self.assertNotIgnored(
                    Request("http://site.local/admin/main"), middleware
                ),
                self.assertNotIgnored(Request("http://site.local/static/"), middleware),
            ],
            fireOnOneErrback=True,
        )

    def test_robotstxt_error(self):
        self.crawler.settings.set("ROBOTSTXT_OBEY", True)
        err = error.DNSLookupError("Robotstxt address not found")

        def return_failure(request):
            reactor.callFromThread(deferred.errback, failure.Failure(err))
            return deferred

        self.crawler.engine.download.side_effect = return_failure

        middleware = RobotsTxtMiddleware(self.crawler)
        middleware._logerror = mock.MagicMock(side_effect=middleware._logerror)
        deferred = middleware.process_request(Request("http://site.local"), None)
        deferred.addCallback(lambda _: self.assertTrue(middleware._logerro.called))