import contextlib
import textwrap

# These flags are used to compile any C++ source within Numpy.
# They are chosen to have very few runtime dependencies.
NPY_CXX_FLAGS = [
    "-std=c++11",  # Minimal standard version
    "-D__STDC_VERSION__=0",  # for compatibility with C headers
    "-fno-exceptions",  # no exception support
    "-fno-rtti"]  # no runtime type information


class _Config:
    """An abstract class holds all configurable attributes of `CCompilerOpt`,
    these class attributes can be used to change the default behavior
    of `CCompilerOpt` in order to fit other requirements.

    Attributes
    ----------
    conf_nocache : bool
        Set True to disable memory and file cache.
        Default is False.

    conf_noopt : bool
        Set True to forces the optimization to be disabled,
        in this case `CCompilerOpt` tends to generate all
        expected headers in order to 'not' break the build.
        Default is False.

    conf_cache_factors : list
        Add extra factors to the primary caching factors. The caching factors
        are utilized to determine if there are changes had happened that
        requires to discard the cache and re-updating it. The primary factors
        are the arguments of `CCompilerOpt` and `CCompiler`'s properties(type, flags, etc).
        Default is list of two items, containing the time of last modification
        of `ccompiler_opt` and value of attribute "conf_noopt"

    conf_tmp_path : str,
        The path of temporary directory. Default is auto-created
        temporary directory via ``tempfile.mkdtemp()``.

    conf_check_path : str
        The path of testing files. Each added CPU feature must have a
        **C** source file contains at least one intrinsic or instruction that
        related to this feature, so it can be tested against the compiler.
        Default is ``./distutils/checks``.

    conf_target_groups : dict
        Extra tokens that can be reached from dispatch-able sources through
        the special mark ``@targets``. Default is an empty dictionary.

        **Notes**:
            - case-insensitive for tokens and group names
            - sign '#' must stick in the begin of group name and only within ``@targets``

        **Example**:
            .. code-block:: console

                $ "@targets #avx_group other_tokens" > group_inside.c

            >>> CCompilerOpt.conf_target_groups["avx_group"] = \\
            "$werror $maxopt avx2 avx512f avx512_skx"
            >>> cco = CCompilerOpt(cc_instance)
            >>> cco.try_dispatch(["group_inside.c"])

    conf_c_prefix : str
        The prefix of public C definitions. Default is ``"NPY_"``.

    conf_c_prefix_ : str
        The prefix of internal C definitions. Default is ``"NPY__"``.

    conf_cc_flags : dict
        Nested dictionaries defining several compiler flags
        that linked to some major functions, the main key
        represent the compiler name and sub-keys represent
        flags names. Default is already covers all supported
        **C** compilers.

        Sub-keys explained as follows:

        "native": str or None
            used by argument option `native`, to detect the current
            machine support via the compiler.
        "werror": str or None
            utilized to treat warning as errors during testing CPU features
            against the compiler and also for target's policy `$werror`
            via dispatch-able sources.
        "maxopt": str or None
            utilized for target's policy '$maxopt' and the value should
            contains the maximum acceptable optimization by the compiler.
            e.g. in gcc `'-O3'`

        **Notes**:
            * case-sensitive for compiler names and flags
            * use space to separate multiple flags
            * any flag will tested against the compiler and it will skipped
              if it's not applicable.

    conf_min_features : dict
        A dictionary defines the used CPU features for
        argument option `'min'`, the key represent the CPU architecture
        name e.g. `'x86'`. Default values provide the best effort
        on wide range of users platforms.

        **Note**: case-sensitive for architecture names.

    conf_features : dict
        Nested dictionaries used for identifying the CPU features.
        the primary key is represented as a feature name or group name
        that gathers several features. Default values covers all
        supported features but without the major options like "flags",
        these undefined options handle it by method `conf_features_partial()`.
        Default value is covers almost all CPU features for *X86*, *IBM/Power64*
        and *ARM 7/8*.

        Sub-keys explained as follows:

        "implies" : str or list, optional,
            List of CPU feature names to be implied by it,
            the feature name must be defined within `conf_features`.
            Default is None.

        "flags": str or list, optional
            List of compiler flags. Default is None.

        "detect": str or list, optional
            List of CPU feature names that required to be detected
            in runtime. By default, its the feature name or features
            in "group" if its specified.

        "implies_detect": bool, optional
            If True, all "detect" of implied features will be combined.
            Default is True. see `feature_detect()`.

        "group": str or list, optional
            Same as "implies" but doesn't require the feature name to be
            defined within `conf_features`.

        "interest": int, required
            a key for sorting CPU features

        "headers": str or list, optional
            intrinsics C header file

        "disable": str, optional
            force disable feature, the string value should contains the
            reason of disabling.

        "autovec": bool or None, optional
            True or False to declare that CPU feature can be auto-vectorized
            by the compiler.
            By default(None), treated as True if the feature contains at
            least one applicable flag. see `feature_can_autovec()`

        "extra_checks": str or list, optional
            Extra test case names for the CPU feature that need to be tested
            against the compiler.

            Each test case must have a C file named ``extra_xxxx.c``, where
            ``xxxx`` is the case name in lower case, under 'conf_check_path'.
            It should contain at least one intrinsic or function related to the test case.

            If the compiler able to successfully compile the C file then `CCompilerOpt`
            will add a C ``#define`` for it into the main dispatch header, e.g.
            ``#define {conf_c_prefix}_XXXX`` where ``XXXX`` is the case name in upper case.

        **NOTES**:
            * space can be used as separator with options that supports "str or list"
            * case-sensitive for all values and feature name must be in upper-case.
            * if flags aren't applicable, its will skipped rather than disable the
              CPU feature
            * the CPU feature will disabled if the compiler fail to compile
              the test file
    """

    conf_nocache = False
    conf_noopt = False
    conf_cache_factors = None
    conf_tmp_path = None
    conf_check_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "checks",
    )
    conf_target_groups = {}
    conf_c_prefix = "NPY_"
    conf_c_prefix_ = "NPY__"
    conf_cc_flags = {
        "gcc": {
            # native should always fail on arm and ppc64,
            # native usually works only with x86
            "native": "-march=native",
            "opt": "-O3",
            "werror": "-Werror",
        },
        "clang": {
            "native": "-march=native",
            "opt": "-O3",
            # One of the following flags needs to be applicable for Clang to
            # guarantee the sanity of the testing process, however in certain
            # cases `-Werror` gets skipped during the availability test due to
            # "unused arguments" warnings.
            # see https://github.com/numpy/numpy/issues/19624
            "werror": "-Werror=switch -Werror",
        },
        "icc": {
            "native": "-xHost",
            "opt": "-O3",
            "werror": "-Werror",
        },
        "iccw": {
            "native": "/QxHost",
            "opt": "/O3",
            "werror": "/Werror",
        },
        "msvc": {
            "native": None,
            "opt": "/O2",
            "werror": "/WX",
        },
    }
    conf_min_features = {
        "x86": "SSE SSE2",
        "x64": "SSE SSE2 SSE3",
        "ppc64": "", # play it safe
        "ppc64le": "VSX VSX2",
        "s390x": "",
        "armhf": "", # play it safe
        "aarch64": "NEON NEON_FP16 NEON_VFPV4 ASIMD",
    }
    conf_features = {
        # X86
        "SSE": {
            "interest": 1, "headers": "xmmintrin.h",
            # enabling SSE without SSE2 is useless also
            # it's non-optional for x86_64
            "implies": "SSE2",
        },
        "SSE2": {"interest": 2, "implies": "SSE", "headers": "emmintrin.h"},
        "SSE3": {"interest": 3, "implies": "SSE2", "headers": "pmmintrin.h"},
        "SSSE3": {"interest": 4, "implies": "SSE3", "headers": "tmmintrin.h"},
        "SSE41": {"interest": 5, "implies": "SSSE3", "headers": "smmintrin.h"},
        "POPCNT": {"interest": 6, "implies": "SSE41", "headers": "popcntintrin.h"},
        "SSE42": {"interest": 7, "implies": "POPCNT"},
        "AVX": {
            "interest": 8, "implies": "SSE42", "headers": "immintrin.h",
            "implies_detect": False,
        },
        "XOP": {"interest": 9, "implies": "AVX", "headers": "x86intrin.h"},
        "FMA4": {"interest": 10, "implies": "AVX", "headers": "x86intrin.h"},
        "F16C": {"interest": 11, "implies": "AVX"},
        "FMA3": {"interest": 12, "implies": "F16C"},
        "AVX2": {"interest": 13, "implies": "F16C"},
        "AVX512F": {
            "interest": 20, "implies": "FMA3 AVX2", "implies_detect": False,
            "extra_checks": "AVX512F_REDUCE",
        },
        "AVX512CD": {"interest": 21, "implies": "AVX512F"},
        "AVX512_KNL": {
            "interest": 40, "implies": "AVX512CD", "group": "AVX512ER AVX512PF",
            "detect": "AVX512_KNL", "implies_detect": False,
        },
        "AVX512_KNM": {
            "interest": 41, "implies": "AVX512_KNL",
            "group": "AVX5124FMAPS AVX5124VNNIW AVX512VPOPCNTDQ",
            "detect": "AVX512_KNM", "implies_detect": False,
        },
        "AVX512_SKX": {
            "interest": 42, "implies": "AVX512CD", "group": "AVX512VL AVX512BW AVX512DQ",
            "detect": "AVX512_SKX", "implies_detect": False,
            "extra_checks": "AVX512BW_MASK AVX512DQ_MASK",
        },
        "AVX512_CLX": {
            "interest": 43, "implies": "AVX512_SKX", "group": "AVX512VNNI",
            "detect": "AVX512_CLX",
        },
        "AVX512_CNL": {
            "interest": 44, "implies": "AVX512_SKX", "group": "AVX512IFMA AVX512VBMI",
            "detect": "AVX512_CNL", "implies_detect": False,
        },
        "AVX512_ICL": {
            "interest": 45, "implies": "AVX512_CLX AVX512_CNL",
            "group": "AVX512VBMI2 AVX512BITALG AVX512VPOPCNTDQ",
            "detect": "AVX512_ICL", "implies_detect": False,
        },
        # IBM/Power
        ## Power7/ISA 2.06
        "VSX": {"interest": 1, "headers": "altivec.h", "extra_checks": "VSX_ASM"},
        ## Power8/ISA 2.07
        "VSX2": {"interest": 2, "implies": "VSX", "implies_detect": False},
        "VSX3": {"interest": 3, "implies": "VSX2", "implies_detect": False},
        ## Power10/ISA 3.1
        "VSX4": {"interest": 4, "implies": "VSX3", "implies_detect": False,
                    "extra_checks": "VSX4_MMA"},
        # IBM/Z
        "VX": {"interest": 1, "headers": "vecintrin.h"},
        ## Vector-Enhancements Facility
        "VXE": {"interest": 2, "implies": "VX", "implies_detect": False},
        ## Vector-Enhancements Facility 2
        "VXE2": {"interest": 3, "implies": "VXE", "implies_detect": False},
        # ARM
        "NEON": {"interest": 1, "headers": "arm_neon.h"},
        "NEON_FP16": {"interest": 2, "implies": "NEON"},
        ## FMA
        "NEON_VFPV4": {"interest": 3, "implies": "NEON_FP16"},
        ## Advanced SIMD
        "ASIMD": {"interest": 4, "implies": "NEON_FP16 NEON_VFPV4", "implies_detect": False},
        ## ARMv8.2 half-precision & vector arithm
        "ASIMDHP": {"interest": 5, "implies": "ASIMD"},
        ## ARMv8.2 dot product
        "ASIMDDP": {"interest": 6, "implies": "ASIMD"},
        ## ARMv8.2 Single & half-precision Multiply
        "ASIMDFHM": {"interest": 7, "implies": "ASIMDHP"},
    }
    def conf_features_partial(self):
        """Return a dictionary of supported CPU features by the platform,
        and accumulate the rest of undefined options in `conf_features`,
        the returned dict has same rules and notes in
        class attribute `conf_features`, also its override
        any options that been set in 'conf_features'.
        """
        if self.cc_noopt:
            # optimization is disabled
            return {}

        on_x86 = self.cc_on_x86 or self.cc_on_x64
        is_unix = self.cc_is_gcc or self.cc_is_clang

        if on_x86 and is_unix: return {
            "SSE": {"flags": "-msse"},
            "SSE2": {"flags": "-msse2"},
            "SSE3": {"flags": "-msse3"},
            "SSSE3": {"flags": "-mssse3"},
            "SSE41": {"flags": "-msse4.1"},
            "POPCNT": {"flags": "-mpopcnt"},
            "SSE42": {"flags": "-msse4.2"},
            "AVX": {"flags": "-mavx"},
            "F16C": {"flags": "-mf16c"},
            "XOP": {"flags": "-mxop"},
            "FMA4": {"flags": "-mfma4"},
            "FMA3": {"flags": "-mfma"},
            "AVX2": {"flags": "-mavx2"},
            "AVX512F": {"flags": "-mavx512f -mno-mmx"},
            "AVX512CD": {"flags": "-mavx512cd"},
            "AVX512_KNL": {"flags": "-mavx512er -mavx512pf"},
            "AVX512_KNM": {
                "flags": "-mavx5124fmaps -mavx5124vnniw -mavx512vpopcntdq",
            },
            "AVX512_SKX": {"flags": "-mavx512vl -mavx512bw -mavx512dq"},
            "AVX512_CLX": {"flags": "-mavx512vnni"},
            "AVX512_CNL": {"flags": "-mavx512ifma -mavx512vbmi"},
            "AVX512_ICL": {
                "flags": "-mavx512vbmi2 -mavx512bitalg -mavx512vpopcntdq",
            },
        }
        if on_x86 and self.cc_is_icc: return {
            "SSE": {"flags": "-msse"},
            "SSE2": {"flags": "-msse2"},
            "SSE3": {"flags": "-msse3"},
            "SSSE3": {"flags": "-mssse3"},
            "SSE41": {"flags": "-msse4.1"},
            "POPCNT": {},
            "SSE42": {"flags": "-msse4.2"},
            "AVX": {"flags": "-mavx"},
            "F16C": {},
            "XOP": {"disable": "Intel Compiler doesn't support it"},
            "FMA4": {"disable": "Intel Compiler doesn't support it"},
            # Intel Compiler doesn't support AVX2 or FMA3 independently
            "FMA3": {
                "implies": "F16C AVX2", "flags": "-march=core-avx2",
            },
            "AVX2": {"implies": "FMA3", "flags": "-march=core-avx2"},
            # Intel Compiler doesn't support AVX512F or AVX512CD independently
            "AVX512F": {
                "implies": "AVX2 AVX512CD", "flags": "-march=common-avx512",
            },
            "AVX512CD": {
                "implies": "AVX2 AVX512F", "flags": "-march=common-avx512",
            },
            "AVX512_KNL": {"flags": "-xKNL"},
            "AVX512_KNM": {"flags": "-xKNM"},
            "AVX512_SKX": {"flags": "-xSKYLAKE-AVX512"},
            "AVX512_CLX": {"flags": "-xCASCADELAKE"},
            "AVX512_CNL": {"flags": "-xCANNONLAKE"},
            "AVX512_ICL": {"flags": "-xICELAKE-CLIENT"},
        }
        if on_x86 and self.cc_is_iccw: return {
            "SSE": {"flags": "/arch:SSE"},
            "SSE2": {"flags": "/arch:SSE2"},
            "SSE3": {"flags": "/arch:SSE3"},
            "SSSE3": {"flags": "/arch:SSSE3"},
            "SSE41": {"flags": "/arch:SSE4.1"},
            "POPCNT": {},
            "SSE42": {"flags": "/arch:SSE4.2"},
            "AVX": {"flags": "/arch:AVX"},
            "F16C": {},
            "XOP": {"disable": "Intel Compiler doesn't support it"},
            "FMA4": {"disable": "Intel Compiler doesn't support it"},
            # Intel Compiler doesn't support FMA3 or AVX2 independently
            "FMA3": {
                "implies": "F16C AVX2", "flags": "/arch:CORE-AVX2",
            },
            "AVX2": {
                "implies": "FMA3", "flags": "/arch:CORE-AVX2",
            },
            # Intel Compiler doesn't support AVX512F or AVX512CD independently
            "AVX512F": {
                "implies": "AVX2 AVX512CD", "flags": "/Qx:COMMON-AVX512",
            },
            "AVX512CD": {
                "implies": "AVX2 AVX512F", "flags": "/Qx:COMMON-AVX512",
            },
            "AVX512_KNL": {"flags": "/Qx:KNL"},
            "AVX512_KNM": {"flags": "/Qx:KNM"},
            "AVX512_SKX": {"flags": "/Qx:SKYLAKE-AVX512"},
            "AVX512_CLX": {"flags": "/Qx:CASCADELAKE"},
            "AVX512_CNL": {"flags": "/Qx:CANNONLAKE"},
            "AVX512_ICL": {"flags": "/Qx:ICELAKE-CLIENT"},
        }
        if on_x86 and self.cc_is_msvc: return {
            "SSE": {"flags": "/arch:SSE"} if self.cc_on_x86 else {},
            "SSE2": {"flags": "/arch:SSE2"} if self.cc_on_x86 else {},
            "SSE3": {},
            "SSE41": {},
            "POPCNT": {"headers": "nmmintrin.h"},
            "SSE42": {},
            "AVX": {"flags": "/arch:AVX"},
            "F16C": {},
            "XOP": {"headers": "ammintrin.h"},
            "FMA4": {"headers": "ammintrin.h"},
            # MSVC doesn't support FMA3 or AVX2 independently
            "FMA3": {
                "implies": "F16C AVX2", "flags": "/arch:AVX2",
            },
            "AVX2": {
                "implies": "F16C FMA3", "flags": "/arch:AVX2",
            },
            # MSVC doesn't support AVX512F or AVX512CD independently,
            # always generate instructions belong to (VL/VW/DQ)
            "AVX512F": {
                "implies": "AVX2 AVX512CD AVX512_SKX", "flags": "/arch:AVX512",
            },
            "AVX512CD": {
                "implies": "AVX512F AVX512_SKX", "flags": "/arch:AVX512",
            },
            "AVX512_KNL": {
                "disable": "MSVC compiler doesn't support it",
            },
            "AVX512_KNM": {
                "disable": "MSVC compiler doesn't support it",
            },
            "AVX512_SKX": {"flags": "/arch:AVX512"},
            "AVX512_CLX": {},
            "AVX512_CNL": {},
            "AVX512_ICL": {},
        }

        on_power = self.cc_on_ppc64le or self.cc_on_ppc64
        if on_power:
            partial = {
                "VSX": {
                    "implies": ("VSX2" if self.cc_on_ppc64le else ""),
                    "flags": "-mvsx",
                },
                "VSX2": {
                    "flags": "-mcpu=power8", "implies_detect": False,
                },
                "VSX3": {
                    "flags": "-mcpu=power9 -mtune=power9", "implies_detect": False,
                },
                "VSX4": {
                    "flags": "-mcpu=power10 -mtune=power10", "implies_detect": False,
                },
            }
            if self.cc_is_clang:
                partial["VSX"]["flags"]  = "-maltivec -mvsx"
                partial["VSX2"]["flags"] = "-mpower8-vector"
                partial["VSX3"]["flags"] = "-mpower9-vector"
                partial["VSX4"]["flags"] = "-mpower10-vector"

            return partial

        on_zarch = self.cc_on_s390x
        if on_zarch:
            partial = {
                "VX": {
                    "flags": "-march=arch11 -mzvector",
                },
                "VXE": {
                    "flags": "-march=arch12", "implies_detect": False,
                },
                "VXE2": {
                    "flags": "-march=arch13", "implies_detect": False,
                },
            }

            return partial


        if self.cc_on_aarch64 and is_unix: return {
            "NEON": {
                "implies": "NEON_FP16 NEON_VFPV4 ASIMD", "autovec": True,
            },
            "NEON_FP16": {
                "implies": "NEON NEON_VFPV4 ASIMD", "autovec": True,
            },
            "NEON_VFPV4": {
                "implies": "NEON NEON_FP16 ASIMD", "autovec": True,
            },
            "ASIMD": {
                "implies": "NEON NEON_FP16 NEON_VFPV4", "autovec": True,
            },
            "ASIMDHP": {
                "flags": "-march=armv8.2-a+fp16",
            },
            "ASIMDDP": {
                "flags": "-march=armv8.2-a+dotprod",
            },
            "ASIMDFHM": {
                "flags": "-march=armv8.2-a+fp16fml",
            },
        }
        if self.cc_on_armhf and is_unix: return {
            "NEON": {
                "flags": "-mfpu=neon",
            },
            "NEON_FP16": {
                "flags": "-mfpu=neon-fp16 -mfp16-format=ieee",
            },
            "NEON_VFPV4": {
                "flags": "-mfpu=neon-vfpv4",
            },
            "ASIMD": {
                "flags": "-mfpu=neon-fp-armv8 -march=armv8-a+simd",
            },
            "ASIMDHP": {
                "flags": "-march=armv8.2-a+fp16",
            },
            "ASIMDDP": {
                "flags": "-march=armv8.2-a+dotprod",
            },
            "ASIMDFHM": {
                "flags": "-march=armv8.2-a+fp16fml",
            },
        }
        # TODO: ARM MSVC
        return {}

    def __init__(self) -> None:
        if self.conf_tmp_path is None:
            import shutil
            import tempfile
            tmp = tempfile.mkdtemp()
            def rm_temp():
                with contextlib.suppress(OSError):
                    shutil.rmtree(tmp)

            atexit.register(rm_temp)
            self.conf_tmp_path = tmp

        if self.conf_cache_factors is None:
            self.conf_cache_factors = [
                os.path.getmtime(__file__),
                self.conf_nocache,
            ]

class _Distutils:
    """A helper class that provides a collection of fundamental methods
    implemented in a top of Python and NumPy Distutils.

    The idea behind this class is to gather all methods that it may
    need to override in case of reuse 'CCompilerOpt' in environment
    different than of what NumPy has.

    Parameters
    ----------
    ccompiler : `CCompiler`
        The generate instance that returned from `distutils.ccompiler.new_compiler()`.
    """

    def __init__(self, ccompiler) -> None:
        self._ccompiler = ccompiler

    def dist_compile(self, sources, flags, ccompiler=None, **kwargs):
        """Wrap CCompiler.compile()."""
        assert(isinstance(sources, list))
        assert(isinstance(flags, list))
        flags = kwargs.pop("extra_postargs", []) + flags
        if not ccompiler:
            ccompiler = self._ccompiler

        return ccompiler.compile(sources, extra_postargs=flags, **kwargs)

    def dist_test(self, source, flags, macros=[]):
        """Return True if 'CCompiler.compile()' able to compile
        a source file with certain flags.
        """
        assert(isinstance(source, str))
        from distutils.errors import CompileError
        cc = self._ccompiler
        bk_spawn = getattr(cc, "spawn", None)
        if bk_spawn:
            cc_type = getattr(self._ccompiler, "compiler_type", "")
            if cc_type in ("msvc",):
                cc.spawn = self._dist_test_spawn_paths
            else:
                cc.spawn = self._dist_test_spawn
        test = False
        try:
            self.dist_compile(
                [source], flags, macros=macros, output_dir=self.conf_tmp_path,
            )
            test = True
        except CompileError as e:
            self.dist_log(str(e), stderr=True)
        if bk_spawn:
            cc.spawn = bk_spawn
        return test

    def dist_info(self):
        """Return a tuple containing info about (platform, compiler, extra_args),
        required by the abstract class '_CCompiler' for discovering the
        platform environment. This is also used as a cache factor in order
        to detect any changes happening from outside.
        """
        if hasattr(self, "_dist_info"):
            return self._dist_info

        cc_type = getattr(self._ccompiler, "compiler_type", "")
        if cc_type in ("intelem", "intelemw"):
            platform = "x86_64"
        elif cc_type in ("intel", "intelw", "intele"):
            platform = "x86"
        else:
            from distutils.util import get_platform
            platform = get_platform()

        cc_info = getattr(self._ccompiler, "compiler", getattr(self._ccompiler, "compiler_so", ""))
        if not cc_type or cc_type == "unix":
            compiler = cc_info[0] if hasattr(cc_info, "__iter__") else str(cc_info)
        else:
            compiler = cc_type

        if hasattr(cc_info, "__iter__") and len(cc_info) > 1:
            extra_args = " ".join(cc_info[1:])
        else:
            extra_args  = os.environ.get("CFLAGS", "")
            extra_args += os.environ.get("CPPFLAGS", "")

        self._dist_info = (platform, compiler, extra_args)
        return self._dist_info

    @staticmethod
    def dist_error(*args):
        """Raise a compiler error."""
        from distutils.errors import CompileError
        raise CompileError(_Distutils._dist_str(*args))

    @staticmethod
    def dist_fatal(*args):
        """Raise a distutils error."""
        from distutils.errors import DistutilsError
        raise DistutilsError(_Distutils._dist_str(*args))

    @staticmethod
    def dist_log(*args, stderr=False):
        """Print a console message."""
        from numpy.distutils import log
        out = _Distutils._dist_str(*args)
        if stderr:
            log.warn(out)
        else:
            log.info(out)

    @staticmethod
    def dist_load_module(name, path):
        """Load a module from file, required by the abstract class '_Cache'."""
        from .misc_util import exec_mod_from_location
        try:
            return exec_mod_from_location(name, path)
        except Exception as e:
            _Distutils.dist_log(e, stderr=True)
        return None

    @staticmethod
    def _dist_str(*args):
        """Return a string to print by log and errors."""
        def to_str(arg):
            if not isinstance(arg, str) and hasattr(arg, "__iter__"):
                ret = []
                for a in arg:
                    ret.append(to_str(a))
                return "("+ " ".join(ret) + ")"
            return str(arg)

        stack = inspect.stack()[2]
        start = "CCompilerOpt.%s[%d] : " % (stack.function, stack.lineno)
        out = " ".join([
            to_str(a)
            for a in (*args,)
        ])
        return start + out

    def _dist_test_spawn_paths(self, cmd, display=None):
        """Fix msvc SDK ENV path same as distutils do
        without it we get c1: fatal error C1356: unable to find mspdbcore.dll.
        """
        if not hasattr(self._ccompiler, "_paths"):
            self._dist_test_spawn(cmd)
            return
        old_path = os.getenv("PATH")
        try:
            os.environ["PATH"] = self._ccompiler._paths
            self._dist_test_spawn(cmd)
        finally:
            os.environ["PATH"] = old_path

    _dist_warn_regex = re.compile(
        # intel and msvc compilers don't raise
        # fatal errors when flags are wrong or unsupported
        ".*("
        "warning D9002|"  # msvc, it should be work with any language.
        "invalid argument for option" # intel
        ").*",
    )
    @staticmethod
    def _dist_test_spawn(cmd, display=None):
        try:
            o = subprocess.check_output(cmd, stderr=subprocess.STDOUT,
                                        universal_newlines=True)
            if o and re.match(_Distutils._dist_warn_regex, o):
                _Distutils.dist_error(
                    "Flags in command", cmd ,"aren't supported by the compiler"
                    ", output -> \n%s" % o,
                )
        except subprocess.CalledProcessError as exc:
            o = exc.output
            s = exc.returncode
        except OSError as e:
            o = e
            s = 127
        else:
            return
        _Distutils.dist_error(
            "Command", cmd, "failed with exit status %d output -> \n%s" % (
            s, o,
        ))

_share_cache = {}
class _Cache:
    """An abstract class handles caching functionality, provides two
    levels of caching, in-memory by share instances attributes among
    each other and by store attributes into files.

    **Note**:
        any attributes that start with ``_`` or ``conf_`` will be ignored.

    Parameters
    ----------
    cache_path : str or None
        The path of cache file, if None then cache in file will disabled.

    *factors :
        The caching factors that need to utilize next to `conf_cache_factors`.

    Attributes
    ----------
    cache_private : set
        Hold the attributes that need be skipped from "in-memory cache".

    cache_infile : bool
        Utilized during initializing this class, to determine if the cache was able
        to loaded from the specified cache path in 'cache_path'.
    """

    # skip attributes from cache
    _cache_ignore = re.compile("^(_|conf_)")

    def __init__(self, cache_path=None, *factors) -> None:
        self.cache_me = {}
        self.cache_private = set()
        self.cache_infile = False
        self._cache_path = None

        if self.conf_nocache:
            self.dist_log("cache is disabled by `Config`")
            return

        self._cache_hash = self.cache_hash(*factors, *self.conf_cache_factors)
        self._cache_path = cache_path
        if cache_path and os.path.exists(cache_path):
            self.dist_log("load cache from file ->", cache_path)
            cache_mod = self.dist_load_module("cache", cache_path)
            if not cache_mod:
                self.dist_log(
                    "unable to load the cache file as a module",
                    stderr=True,
                )
            elif not hasattr(cache_mod, "hash") or \
                     not hasattr(cache_mod, "data"):
                self.dist_log("invalid cache file", stderr=True)
            elif self._cache_hash == cache_mod.hash:
                self.dist_log("hit the file cache")
                for attr, val in cache_mod.data.items():
                    setattr(self, attr, val)
                self.cache_infile = True
            else:
                self.dist_log("miss the file cache")

        if not self.cache_infile:
            other_cache = _share_cache.get(self._cache_hash)
            if other_cache:
                self.dist_log("hit the memory cache")
                for attr, val in other_cache.__dict__.items():
                    if attr in other_cache.cache_private or \
                               re.match(self._cache_ignore, attr):
                        continue
                    setattr(self, attr, val)

        _share_cache[self._cache_hash] = self
        atexit.register(self.cache_flush)

    def __del__(self) -> None:
        for h, o in _share_cache.items():
            if o == self:
                _share_cache.pop(h)
                break

    def cache_flush(self):
        """Force update the cache."""
        if not self._cache_path:
            return
        # TODO: don't write if the cache doesn't change
        self.dist_log("write cache to path ->", self._cache_path)
        cdict = self.__dict__.copy()
        for attr in self.__dict__:
            if re.match(self._cache_ignore, attr):
                cdict.pop(attr)

        d = os.path.dirname(self._cache_path)
        if not os.path.exists(d):
            os.makedirs(d)

        repr_dict = pprint.pformat(cdict, compact=True)
        with open(self._cache_path, "w") as f:
            f.write(textwrap.dedent("""\
            # AUTOGENERATED DON'T EDIT
            # Please make changes to the code generator \
            (distutils/ccompiler_opt.py)
            hash = {}
            data = \\
            """).format(self._cache_hash))
            f.write(repr_dict)

    def cache_hash(self, *factors):
        # is there a built-in non-crypto hash?
        # sdbm
        chash = 0
        for f in factors:
            for char in str(f):
                chash  = ord(char) + (chash << 6) + (chash << 16) - chash
                chash &= 0xFFFFFFFF
        return chash

    @staticmethod
    def me(cb):
        """A static method that can be treated as a decorator to
        dynamically cache certain methods.
        """
        def cache_wrap_me(self, *args, **kwargs):
            # good for normal args
            cache_key = str((
                cb.__name__, *args, *kwargs.keys(), *kwargs.values(),
            ))
            if cache_key in self.cache_me:
                return self.cache_me[cache_key]
            ccb = cb(self, *args, **kwargs)
            self.cache_me[cache_key] = ccb
            return ccb
        return cache_wrap_me

class _CCompiler:
    """A helper class for `CCompilerOpt` containing all utilities that
    related to the fundamental compiler's functions.

    Attributes
    ----------
    cc_on_x86 : bool
        True when the target architecture is 32-bit x86
    cc_on_x64 : bool
        True when the target architecture is 64-bit x86
    cc_on_ppc64 : bool
        True when the target architecture is 64-bit big-endian powerpc
    cc_on_ppc64le : bool
        True when the target architecture is 64-bit litle-endian powerpc
    cc_on_s390x : bool
        True when the target architecture is IBM/ZARCH on linux
    cc_on_armhf : bool
        True when the target architecture is 32-bit ARMv7+
    cc_on_aarch64 : bool
        True when the target architecture is 64-bit Armv8-a+
    cc_on_noarch : bool
        True when the target architecture is unknown or not supported
    cc_is_gcc : bool
        True if the compiler is GNU or
        if the compiler is unknown
    cc_is_clang : bool
        True if the compiler is Clang
    cc_is_icc : bool
        True if the compiler is Intel compiler (unix like)
    cc_is_iccw : bool
        True if the compiler is Intel compiler (msvc like)
    cc_is_nocc : bool
        True if the compiler isn't supported directly,
        Note: that cause a fail-back to gcc
    cc_has_debug : bool
        True if the compiler has debug flags
    cc_has_native : bool
        True if the compiler has native flags
    cc_noopt : bool
        True if the compiler has definition 'DISABLE_OPT*',
        or 'cc_on_noarch' is True
    cc_march : str
        The target architecture name, or "unknown" if
        the architecture isn't supported
    cc_name : str
        The compiler name, or "unknown" if the compiler isn't supported
    cc_flags : dict
        Dictionary containing the initialized flags of `_Config.conf_cc_flags`
    """

    def __init__(self) -> None:
        if hasattr(self, "cc_is_cached"):
            return
        #      attr            regex        compiler-expression
        detect_arch = (
            ("cc_on_x64",      ".*(x|x86_|amd)64.*", ""),
            ("cc_on_x86",      ".*(win32|x86|i386|i686).*", ""),
            ("cc_on_ppc64le",  ".*(powerpc|ppc)64(el|le).*", ""),
            ("cc_on_ppc64",    ".*(powerpc|ppc)64.*", ""),
            ("cc_on_aarch64",  ".*(aarch64|arm64).*", ""),
            ("cc_on_armhf",    ".*arm.*", "defined(__ARM_ARCH_7__) || "
                                          "defined(__ARM_ARCH_7A__)"),
            ("cc_on_s390x",    ".*s390x.*", ""),
            # undefined platform
            ("cc_on_noarch",   "", ""),
        )
        detect_compiler = (
            ("cc_is_gcc",     r".*(gcc|gnu\-g).*", ""),
            ("cc_is_clang",    ".*clang.*", ""),
            # intel msvc like
            ("cc_is_iccw",     ".*(intelw|intelemw|iccw).*", ""),
            ("cc_is_icc",      ".*(intel|icc).*", ""),  # intel unix like
            ("cc_is_msvc",     ".*msvc.*", ""),
            # undefined compiler will be treat it as gcc
            ("cc_is_nocc",     "", ""),
        )
        detect_args = (
           ("cc_has_debug",  ".*(O0|Od|ggdb|coverage|debug:full).*", ""),
           ("cc_has_native", ".*(-march=native|-xHost|/QxHost).*", ""),
           # in case if the class run with -DNPY_DISABLE_OPTIMIZATION
           ("cc_noopt", ".*DISABLE_OPT.*", ""),
        )

        dist_info = self.dist_info()
        platform, compiler_info, extra_args = dist_info
        # set False to all attrs
        for section in (detect_arch, detect_compiler, detect_args):
            for attr, _rgex, _cexpr in section:
                setattr(self, attr, False)

        for detect, searchin in ((detect_arch, platform), (detect_compiler, compiler_info)):
            for attr, rgex, cexpr in detect:
                if rgex and not re.match(rgex, searchin, re.IGNORECASE):
                    continue
                if cexpr and not self.cc_test_cexpr(cexpr):
                    continue
                setattr(self, attr, True)
                break

        for attr, rgex, cexpr in detect_args:
            if rgex and not re.match(rgex, extra_args, re.IGNORECASE):
                continue
            if cexpr and not self.cc_test_cexpr(cexpr):
                continue
            setattr(self, attr, True)

        if self.cc_on_noarch:
            self.dist_log(
                "unable to detect CPU architecture which lead to disable the optimization. "
                f"check dist_info:<<\n{dist_info}\n>>",
                stderr=True,
            )
            self.cc_noopt = True

        if self.conf_noopt:
            self.dist_log("Optimization is disabled by the Config", stderr=True)
            self.cc_noopt = True

        if self.cc_is_nocc:
            """
            mingw can be treated as a gcc, and also xlc even if it based on clang,
            but still has the same gcc optimization flags.
            """
            self.dist_log(
                "unable to detect compiler type which leads to treating it as GCC. "
                "this is a normal behavior if you're using gcc-like compiler such as MinGW or IBM/XLC."
                f"check dist_info:<<\n{dist_info}\n>>",
                stderr=True,
            )
            self.cc_is_gcc = True

        self.cc_march = "unknown"
        for arch in ("x86", "x64", "ppc64", "ppc64le",
                     "armhf", "aarch64", "s390x"):
            if getattr(self, "cc_on_" + arch):
                self.cc_march = arch
                break

        self.cc_name = "unknown"
        for name in ("gcc", "clang", "iccw", "icc", "msvc"):
            if getattr(self, "cc_is_" + name):
                self.cc_name = name
                break

        self.cc_flags = {}
        compiler_flags = self.conf_cc_flags.get(self.cc_name)
        if compiler_flags is None:
            self.dist_fatal(
                "undefined flag for compiler '%s', "
                "leave an empty dict instead" % self.cc_name,
            )
        for name, flags in compiler_flags.items():
            self.cc_flags[name] = nflags = []
            if flags:
                assert(isinstance(flags, str))
                flags = flags.split()
                for f in flags:
                    if self.cc_test_flags([f]):
                        nflags.append(f)

        self.cc_is_cached = True

    @_Cache.me
    def cc_test_flags(self, flags):
        """Returns True if the compiler supports 'flags'."""
        assert(isinstance(flags, list))
        self.dist_log("testing flags", flags)
        test_path = os.path.join(self.conf_check_path, "test_flags.c")
        test = self.dist_test(test_path, flags)
        if not test:
            self.dist_log("testing failed", stderr=True)
        return test

    @_Cache.me
    def cc_test_cexpr(self, cexpr, flags=[]):
        """Same as the above but supports compile-time expressions."""
        self.dist_log("testing compiler expression", cexpr)
        test_path = os.path.join(self.conf_tmp_path, "npy_dist_test_cexpr.c")
        with open(test_path, "w") as fd:
            fd.write(textwrap.dedent(f"""\
               #if !({cexpr})
                   #error "unsupported expression"
               #endif
               int dummy;
            """))
        test = self.dist_test(test_path, flags)
        if not test:
            self.dist_log("testing failed", stderr=True)
        return test

    def cc_normalize_flags(self, flags):
        """Remove the conflicts that caused due gathering implied features flags.

        Parameters
        ----------
        'flags' list, compiler flags
            flags should be sorted from the lowest to the highest interest.

        Returns
        -------
        list, filtered from any conflicts.

        Examples
        --------
        >>> self.cc_normalize_flags(['-march=armv8.2-a+fp16', '-march=armv8.2-a+dotprod'])
        ['armv8.2-a+fp16+dotprod']

        >>> self.cc_normalize_flags(
            ['-msse', '-msse2', '-msse3', '-mssse3', '-msse4.1', '-msse4.2', '-mavx', '-march=core-avx2']
        )
        ['-march=core-avx2']
        """
        assert(isinstance(flags, list))
        if self.cc_is_gcc or self.cc_is_clang or self.cc_is_icc:
            return self._cc_normalize_unix(flags)

        if self.cc_is_msvc or self.cc_is_iccw:
            return self._cc_normalize_win(flags)
        return flags

    _cc_normalize_unix_mrgx = re.compile(
        # 1- to check the highest of
        r"^(-mcpu=|-march=|-x[A-Z0-9\-])",
    )
    _cc_normalize_unix_frgx = re.compile(
        # 2- to remove any flags starts with
        # -march, -mcpu, -x(INTEL) and '-m' without '='
        r"^(?!(-mcpu=|-march=|-x[A-Z0-9\-]|-m[a-z0-9\-\.]*.$))|"
        # exclude:
        r"(?:-mzvector)",
    )
    _cc_normalize_unix_krgx = re.compile(
        # 3- keep only the highest of
        r"^(-mfpu|-mtune)",
    )
    _cc_normalize_arch_ver = re.compile(
        r"[0-9.]",
    )
    def _cc_normalize_unix(self, flags):
        def ver_flags(f):
            #        arch ver  subflag
            # -march=armv8.2-a+fp16fml
            tokens = f.split("+")
            ver = float("0" + "".join(
                re.findall(self._cc_normalize_arch_ver, tokens[0]),
            ))
            return ver, tokens[0], tokens[1:]

        if len(flags) <= 1:
            return flags
        # get the highest matched flag
        for i, cur_flag in enumerate(reversed(flags)):
            if not re.match(self._cc_normalize_unix_mrgx, cur_flag):
                continue
            lower_flags = flags[:-(i+1)]
            upper_flags = flags[-i:]
            filterd = list(filter(
                self._cc_normalize_unix_frgx.search, lower_flags,
            ))
            # gather subflags
            ver, arch, subflags = ver_flags(cur_flag)
            if ver > 0 and len(subflags) > 0:
                for xflag in lower_flags:
                    xver, _, xsubflags = ver_flags(xflag)
                    if ver == xver:
                        subflags = xsubflags + subflags
                cur_flag = arch + "+" + "+".join(subflags)

            flags = [*filterd, cur_flag]
            if i > 0:
                flags += upper_flags
            break

        # to remove overridable flags
        final_flags = []
        matched = set()
        for f in reversed(flags):
            match = re.match(self._cc_normalize_unix_krgx, f)
            if not match:
                pass
            elif match[0] in matched:
                continue
            else:
                matched.add(match[0])
            final_flags.insert(0, f)
        return final_flags

    _cc_normalize_win_frgx = re.compile(
        r"^(?!(/arch\:|/Qx\:))",
    )
    _cc_normalize_win_mrgx = re.compile(
        r"^(/arch|/Qx:)",
    )
    def _cc_normalize_win(self, flags):
        for i, f in enumerate(reversed(flags)):
            if not re.match(self._cc_normalize_win_mrgx, f):
                continue
            i += 1
            return list(filter(
                self._cc_normalize_win_frgx.search, flags[:-i],
            )) + flags[-i:]
        return flags

class _Feature:
    """A helper class for `CCompilerOpt` that managing CPU features.

    Attributes
    ----------
    feature_supported : dict
        Dictionary containing all CPU features that supported
        by the platform, according to the specified values in attribute
        `_Config.conf_features` and `_Config.conf_features_partial()`

    feature_min : set
        The minimum support of CPU features, according to
        the specified values in attribute `_Config.conf_min_features`.
    """

    def __init__(self) -> None:
        if hasattr(self, "feature_is_cached"):
            return
        self.feature_supported = pfeatures = self.conf_features_partial()
        for feature_name in list(pfeatures.keys()):
            feature  = pfeatures[feature_name]
            cfeature = self.conf_features[feature_name]
            feature.update({
                k:v for k,v in cfeature.items() if k not in feature
            })
            disabled = feature.get("disable")
            if disabled is not None:
                pfeatures.pop(feature_name)
                self.dist_log(
                    "feature '%s' is disabled," % feature_name,
                    disabled, stderr=True,
                )
                continue
            # list is used internally for these options
            for option in (
                "implies", "group", "detect", "headers", "flags", "extra_checks",
            ) :
                oval = feature.get(option)
                if isinstance(oval, str):
                    feature[option] = oval.split()

        self.feature_min = set()
        min_f = self.conf_min_features.get(self.cc_march, "")
        for F in min_f.upper().split():
            if F in self.feature_supported:
                self.feature_min.add(F)

        self.feature_is_cached = True

    def feature_names(self, names=None, force_flags=None, macros=[]):
        """Returns a set of CPU feature names that supported by platform and the **C** compiler.

        Parameters
        ----------
        names : sequence or None, optional
            Specify certain CPU features to test it against the **C** compiler.
            if None(default), it will test all current supported features.
            **Note**: feature names must be in upper-case.

        force_flags  list or None, optional
            If None(default), default compiler flags for every CPU feature will
            be used during the test.

        macros : list of tuples, optional
            A list of C macro definitions.
        """
        assert(
            names is None or (
                not isinstance(names, str) and
                hasattr(names, "__iter__")
            )
        )
        assert(force_flags is None or isinstance(force_flags, list))
        if names is None:
            names = self.feature_supported.keys()
        supported_names = set()
        for f in names:
            if self.feature_is_supported(
                f, force_flags=force_flags, macros=macros,
            ):
                supported_names.add(f)
        return supported_names

    def feature_is_exist(self, name):
        """Returns True if a certain feature is exist and covered within
        `_Config.conf_features`.

        Parameters
        ----------
        'name': str
            feature name in uppercase.
        """
        assert(name.isupper())
        return name in self.conf_features

    def feature_sorted(self, names, reverse=False):
        """Sort a list of CPU features ordered by the lowest interest.

        Parameters
        ----------
        'names': sequence
            sequence of supported feature names in uppercase.
        'reverse': bool, optional
            If true, the sorted features is reversed. (highest interest)

        Returns
        -------
        list, sorted CPU features
        """
        def sort_cb(k):
            if isinstance(k, str):
                return self.feature_supported[k]["interest"]
            # multiple features
            rank = max([self.feature_supported[f]["interest"] for f in k])
            # FIXME: that's not a safe way to increase the rank for
            # multi targets
            rank += len(k) -1
            return rank
        return sorted(names, reverse=reverse, key=sort_cb)

    def feature_implies(self, names, keep_origins=False):
        """Return a set of CPU features that implied by 'names'.

        Parameters
        ----------
        names : str or sequence of str
            CPU feature name(s) in uppercase.

        keep_origins : bool
            if False(default) then the returned set will not contain any
            features from 'names'. This case happens only when two features
            imply each other.

        Examples
        --------
        >>> self.feature_implies("SSE3")
        {'SSE', 'SSE2'}
        >>> self.feature_implies("SSE2")
        {'SSE'}
        >>> self.feature_implies("SSE2", keep_origins=True)
        # 'SSE2' found here since 'SSE' and 'SSE2' imply each other
        {'SSE', 'SSE2'}
        """
        def get_implies(name, _caller=set()):
            implies = set()
            d = self.feature_supported[name]
            for i in d.get("implies", []):
                implies.add(i)
                if i in _caller:
                    # infinity recursive guard since
                    # features can imply each other
                    continue
                _caller.add(name)
                implies = implies.union(get_implies(i, _caller))
            return implies

        if isinstance(names, str):
            implies = get_implies(names)
            names = [names]
        else:
            assert(hasattr(names, "__iter__"))
            implies = set()
            for n in names:
                implies = implies.union(get_implies(n))
        if not keep_origins:
            implies.difference_update(names)
        return implies

    def feature_implies_c(self, names):
        """Same as feature_implies() but combining 'names'."""
        names = {names} if isinstance(names, str) else set(names)
        return names.union(self.feature_implies(names))

    def feature_ahead(self, names):
        """Return list of features in 'names' after remove any
        implied features and keep the origins.

        Parameters
        ----------
        'names': sequence
            sequence of CPU feature names in uppercase.

        Returns
        -------
        list of CPU features sorted as-is 'names'

        Examples
        --------
        >>> self.feature_ahead(["SSE2", "SSE3", "SSE41"])
        ["SSE41"]
        # assume AVX2 and FMA3 implies each other and AVX2
        # is the highest interest
        >>> self.feature_ahead(["SSE2", "SSE3", "SSE41", "AVX2", "FMA3"])
        ["AVX2"]
        # assume AVX2 and FMA3 don't implies each other
        >>> self.feature_ahead(["SSE2", "SSE3", "SSE41", "AVX2", "FMA3"])
        ["AVX2", "FMA3"]
        """
        assert(
            not isinstance(names, str)
            and hasattr(names, "__iter__")
        )
        implies = self.feature_implies(names, keep_origins=True)
        ahead = [n for n in names if n not in implies]
        if len(ahead) == 0:
            # return the highest interested feature
            # if all features imply each other
            ahead = self.feature_sorted(names, reverse=True)[:1]
        return ahead

    def feature_untied(self, names):
        """Same as 'feature_ahead()' but if both features implied each other
        and keep the highest interest.

        Parameters
        ----------
        'names': sequence
            sequence of CPU feature names in uppercase.

        Returns
        -------
        list of CPU features sorted as-is 'names'

        Examples
        --------
        >>> self.feature_untied(["SSE2", "SSE3", "SSE41"])
        ["SSE2", "SSE3", "SSE41"]
        # assume AVX2 and FMA3 implies each other
        >>> self.feature_untied(["SSE2", "SSE3", "SSE41", "FMA3", "AVX2"])
        ["SSE2", "SSE3", "SSE41", "AVX2"]
        """
        assert(
            not isinstance(names, str)
            and hasattr(names, "__iter__")
        )
        final = []
        for n in names:
            implies = self.feature_implies(n)
            tied = [
                nn for nn in final
                if nn in implies and n in self.feature_implies(nn)
            ]
            if tied:
                tied = self.feature_sorted([*tied, n])
                if n not in tied[1:]:
                    continue
